/*  g_posix_queues.c - POSIX IPC QUEUES TEST
 *  
 *  (C) Copyright XII 2004 Michal 'GiM' Spadlinski <gim at skrzynka dot pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License Version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "g_posix_queue.h"
#include "dynstuff.h"
#include "g_ipc_wspolne.h"

# ifdef HAS_POSIX_QUEUES

extern const int g_data_format_s;
extern struct g_data_format parentp[2], childp[2];
int showtimes(char *str, list_t p);

#define G_NAME "/gimusia"

int prefork_posix_queue(void)
{
	mq_unlink(G_NAME);
	return 0;
}

int child_posix_queue(int off)
{
	mqd_t queue;
	char *p;
	int i, len;
	struct timeval s, k;
	list_t sending = NULL, receiving = NULL;

	for (i=0; i<MAX_OPEN_TRIES; i++)
	{
		if ((queue = mq_open(G_NAME, O_EXCL|O_RDWR)) != (mqd_t)-1)
			break;
		SLEEP(s, 500);
	}
	if (i == MAX_OPEN_TRIES)
	{
		perror ("child: opening queue");
		return -1;
	}
	DEB("[%d] child queue "G_NAME" opened in rw mode fd:%d %d\n", getpid(), queue, i);

	childp[off+0].type = 1;
	childp[off+0].id = getpid();
	p = childp[off+0].p;
	for (i=0; i<IPCQ_S; i++)
		*p++ = 'A';

	DEB("[%d] child queue data prepared\n", getpid());

	for (i=0; i<MAX_QUERIES; i++)
	{
		TIMESTAMP(s);
		DEB("[%d] child sending data\n", getpid());
		mq_send(queue, (char *)&(childp[off+0]), g_data_format_s, 0);		
		DEB("[%d] child data sent to %d\n", getpid(), queue);
		TIMESTAMP(k);

		DIFFTIME(k, s);
		list_add(&sending, &k, sizeof(struct timeval));

		TIMESTAMP(s);
		DEB("[%d] child waiting....\n", getpid());
		len = mq_receive(queue, (char *)&(childp[off+1]), g_data_format_s, NULL);
		TIMESTAMP(k);
		DIFFTIME(k, s);
		list_add(&receiving, &k, sizeof(struct timeval));

		childp[off+1].p[len - 2*long_s] = 0; 
		DEB("[%d] child received data from %li\n", getpid(), childp[off+1].id);

	}
	showtimes("[+] posx msg queue   sending", sending);
	showtimes("[+] posx msg queue receiving", receiving);

	mq_close(queue);

	list_destroy(sending, 1);
	list_destroy(receiving, 1);

	return 0;
}

int parent_posix_queue(int pid)
{
	mqd_t queue;
	int len;
	struct mq_attr p;
	struct timespec tmout;

	p.mq_maxmsg = 3;
	p.mq_msgsize = g_data_format_s;

	if ((queue = mq_open(G_NAME, O_EXCL|O_RDWR|O_CREAT, 0600, &p)) == (mqd_t)-1)
	{
		perror ("parent creating queue");
		return -1;
	}
#ifdef USE_NONBLOCKING_POSIX_QUEUE
	mq_getattr(queue, &p);
	p.mq_flags = O_NONBLOCK;
	mq_setattr(queue, &p, NULL);
	mq_getattr(queue, &p);
#endif
	DEB("[%d] parent queue "G_NAME" created fd: %d\n", getpid(), queue);

	while (1)
	{
		DEB("[%d] parent queue waiting for data on fd: %d\n", getpid(), queue);
		tmout.tv_sec = time(NULL) + 2;
		tmout.tv_nsec = 0;
#ifdef USE_NONBLOCKING_POSIX_QUEUE
		errno = 0;
		if ((len=mq_receive(queue, (char *)&(parentp[0].type), g_data_format_s, NULL)) == -1)
		{
			if (errno == EAGAIN)
#else
		if ((len=mq_timedreceive(queue, (char *)&(parentp[0].type), g_data_format_s, NULL, &tmout)) == -1)
		{
#endif
			if (waitpid(pid, NULL, WNOHANG))
				break;
		} else {
			DEB("[%d] parent queue data received from %li, %d byte[s]\n", getpid(), parentp[0].id, len);

			parentp[0].p[len-2*long_s] = '\0';
			parentp[0].type = parentp[0].id;
			parentp[0].id = getpid();
			mq_send(queue, (char *)&(parentp[0]), g_data_format_s, 1);
		}
	}

	DEB("[%d] parent child exited\n", getpid());

	mq_close(queue);
	mq_unlink(G_NAME);

	return 0;
}

# endif /* HAS_POSIX_QUEUES */
